import numpy as np
from scipy import special
import cvxpy as cp

from walk_around_grid import WalkAroundGrid
from tabular_softmax import TabularSoftmax
from collections import namedtuple
# import matplotlib.pyplot as plt
from typing import Callable

import torch
import time
import tqdm

from multiprocessing import pool
import os
os.system("taskset -p 0xff %d" % os.getpid())
seed = 0
Transition = namedtuple('Transition', ('state', 'action', 'reward', 'cost', 'next_state', 'terminal'))

def get_batch(env, policy, steps, deterministic=False, max_episode_steps=250):
    offline_data = []
    expected_return = []
    expected_cost = []
    per_step_reward = []
    per_step_cost = []

    trajectory = []
    env.reset()
    obs = env.state
    done = False
    episode_steps = 0
    discounted_return = 0.0
    total_cost = 0.0
    for i in range(steps):
        if done:
            expected_return.append(discounted_return)
            expected_cost.append(total_cost)
            per_step_reward.append(discounted_return / np.sum(env.gamma ** np.arange(0,episode_steps,1)))
            per_step_cost.append(total_cost / np.sum(env.gamma ** np.arange(0,episode_steps,1)))
            offline_data.append(trajectory)

            trajectory = []
            env.reset()
            done = False
            obs = env.state
            episode_steps = 0
            discounted_return = 0.0
            total_cost = 0.0

        action = np.argmax(policy.getActionProbabilities(obs)) if deterministic else policy.samplAction(obs)
        next_obs, reward, cost, done = env.step(action)
        done = done or (episode_steps >= max_episode_steps)
        trajectory.append(Transition(obs,action,reward,cost,next_obs,done))
        obs = next_obs
        discounted_return += ((env.gamma ** episode_steps) * reward)
        total_cost += ((env.gamma ** episode_steps) * cost)
        episode_steps += 1
          
    return offline_data, np.array(expected_return).mean(), np.array(expected_cost).mean(), np.array(per_step_reward).mean(), np.array(per_step_cost).mean()

class TabularCOPO():
    def __init__(self, shape:tuple, data:list, dice_data, gamma:float, alpha:float, beta:float, f:Callable, steps:int):
        self.num_states = shape[0]
        self.num_actions = shape[1]
        self.data = data
        self.gamma = gamma
        self.alpha = alpha
        self.beta = beta
        self.f = f
        self.steps = steps
        mu, data_dist, emp_r, emp_t, emp_c = self.compute_visitation(data)
        self.mu = mu
        self.data_dist = data_dist
        self.empirical_reward = emp_r
        self.empirical_transition = emp_t
        self.empirical_cost = emp_c

    def compute_visitation(self, data):
        '''
            This function accepts a list of transitions and outputs various
            statistics of the data.
        '''
        init_state_density = np.zeros(self.num_states)
        offline_visitation_density = np.zeros((self.num_states,self.num_actions))
        empirical_reward = np.zeros((self.num_states,self.num_actions))
        empirical_cost = np.zeros((self.num_states,self.num_actions))
        empirical_transition = np.zeros((self.num_states,self.num_actions,self.num_states))
        total_steps = 0

        for trajectory in data:
            init_state_density[trajectory[0].state] += 1
            for transition in trajectory:
                offline_visitation_density[transition.state, transition.action] += 1
                empirical_transition[transition.state, transition.action, transition.next_state] += 1
                empirical_reward[transition.state, transition.action] += transition.reward
                empirical_cost[transition.state, transition.action] += transition.cost
                total_steps += 1

        mask = (offline_visitation_density > 0)
        init_state_density = init_state_density / len(data)
        empirical_reward[mask] = empirical_reward[mask] / offline_visitation_density[mask]
        empirical_cost[mask] = empirical_cost[mask] / offline_visitation_density[mask]
        offline_visitation_density = offline_visitation_density / total_steps
        transition_counts = np.expand_dims(empirical_transition.sum(axis=-1), -1)
        transition_counts[transition_counts == 0] += 1
        empirical_transition = np.true_divide(empirical_transition, transition_counts)

        return init_state_density, offline_visitation_density, empirical_reward, empirical_transition, empirical_cost

    def run_policy_opt(self):
        '''
            This function implements a tabular version of AlgaeDICE
        '''
        policy = (1. / self.num_actions) * np.random.randn(self.num_states,self.num_actions)
        value_fn = np.zeros((self.num_states,self.num_actions))
        zeta_fn = np.zeros((self.num_states,self.num_actions))
        lam_value = np.array([0.0])

        lr = 0.01

        lam = torch.tensor(lam_value.astype(np.float32), requires_grad=True)
        v = torch.tensor(value_fn.astype(np.float32), requires_grad=True)
        optimizer1 = torch.optim.Adam([v, lam], lr=lr)#, momentum=0.9)

        zeta = torch.tensor(zeta_fn.astype(np.float32), requires_grad=True)
        pi = torch.tensor(policy.astype(np.float32), requires_grad=True)
        optimizer = torch.optim.Adam([pi, zeta], lr=lr)#, momentum=0.9)

        logB = 0.0
        logV = 0.0
        bcc = 0.0
        bcv = 0.0
        cst_est = 0.0
        obC = 0.0
        for i in range(0,1000):
            lam.data = torch.tensor(lam_value.astype(np.float32), requires_grad=True)
            v.data = torch.tensor(value_fn.astype(np.float32), requires_grad=True)
            br = np.zeros(v.shape)
            for _ in range(self.steps):
                optimizer1.zero_grad()

                a = self.alpha
                d = (1. - self.gamma)

                softmax_policy = torch.tensor(special.softmax(policy, axis=1).astype(np.float32), requires_grad=False)
                spv = torch.mul(softmax_policy, v)
                init_bellman = torch.matmul(spv, torch.ones((self.num_actions,1)))
                first_expectation = d * torch.matmul(init_bellman.T, torch.FloatTensor(self.mu)) if self.gamma < 1.0 else lam

                # this should be |S|x|A|
                # reward should be |S|x|A| and transition should be |S|x|A|x|S|
                static_init_bellman = torch.matmul(torch.mul(softmax_policy, torch.FloatTensor(value_fn)), torch.ones((self.num_actions,1)))
                bellman_residual = self.gamma * torch.stack([(torch.FloatTensor(self.empirical_transition[s,:,:]) @ static_init_bellman).T for s in range(0, self.num_states)]).squeeze()

                bellman_residual += torch.FloatTensor(self.empirical_reward)
                bellman_residual -= v

                second_expectation = a * torch.sum(torch.mul(torch.FloatTensor(self.data_dist), 0.5 * torch.pow(bellman_residual / a, 2))) if self.gamma < 1.0 \
                                        else torch.sum(torch.mul(torch.FloatTensor(self.data_dist), torch.mul(torch.FloatTensor(zeta_fn), -lam + bellman_residual)))

                objective = first_expectation + second_expectation
                if not(self.gamma < 1.0):
                    objective -= a * torch.sum(torch.mul(torch.FloatTensor(self.data_dist), 0.5 * torch.pow(torch.FloatTensor(zeta_fn), 2.0)))
                objective.backward()
                optimizer1.step()

            value_fn = v.detach().numpy()
            lam_value = lam.detach().numpy()

            # solve the outer problem
            zeta.data = torch.tensor(zeta_fn.astype(np.float32), requires_grad=True)
            pi.data = torch.tensor(policy.astype(np.float32), requires_grad=True)
            for _ in range(self.steps):
                optimizer.zero_grad()

                softmax_policy = torch.nn.functional.softmax(pi, dim=1)

                max_init_bellmanV = torch.mul(softmax_policy, torch.FloatTensor(value_fn)) @ torch.ones((self.num_actions,1), dtype=torch.float32)
                max_first_expectationV = d * torch.matmul(torch.t(max_init_bellmanV), torch.FloatTensor(self.mu)) if self.gamma < 1.0 else torch.FloatTensor(lam_value)

                # this should be |S|x|A|
                # reward should be |S|x|A| and transition should be |S|x|A|x|S|
                max_bellman_residualV = (self.gamma * torch.stack([(torch.FloatTensor(self.empirical_transition[s,:,:]) @ max_init_bellmanV).T for s in range(0, self.num_states)], dim=1)).squeeze()
                max_bellman_residualV += torch.FloatTensor(self.empirical_reward) - torch.FloatTensor(value_fn)

                max_second_expectationV = torch.sum(a * torch.mul(torch.FloatTensor(self.data_dist), max_bellman_residualV)) if self.gamma < 1.0 \
                                        else torch.sum(torch.mul(torch.FloatTensor(self.data_dist), torch.mul(zeta, -torch.FloatTensor(lam_value) + max_bellman_residualV)))

                max_objective = -(max_first_expectationV + max_second_expectationV)

                if not(self.gamma < 1.0):
                    max_objective += a * torch.sum(torch.mul(torch.FloatTensor(self.data_dist), 0.5 * torch.pow(zeta, 2.0)))

                max_objective.backward()
                optimizer.step()

            policy = pi.detach().numpy()
            zeta_fn = zeta.detach().numpy()

        return policy, value_fn, zeta_fn, lam_value
    
    def cost_projection_wass_entr_torch(self, policy_params, ref_dist, init_penalty, cost_lim):
        '''
            This function performs a policy projection routine
        '''
        policy = policy_params
        zeta_fn = np.zeros((self.num_states,self.num_actions))
        lam_value = np.array([0.0])
        nu_value = np.array([0.0])

        v = np.zeros((self.num_states,self.num_actions))
        f = np.zeros((self.num_states,self.num_actions))
        lv = 0.0
        nv = 0.00
        f_norm = 0.0
        lr = 0.1
        pi = torch.tensor(policy.astype(np.float32), requires_grad=True)
        optimizer = torch.optim.Adam([pi], lr=lr)#, momentum=0.9)

        value_fn = torch.tensor(v.astype(np.float32), requires_grad=True)
        f_fn = torch.tensor(f.astype(np.float32), requires_grad=True)
        lam = torch.tensor(np.array([lv]).astype(np.float32), requires_grad=True)
        nu = torch.tensor(np.array([nv]).astype(np.float32), requires_grad=True)
        optimizer1 = torch.optim.Adam([value_fn, f_fn, lam, nu], lr=lr)

        # construct Wasserstein transport cost
        action_dist = np.zeros((4,4))
        for n in range(0,4):
            for m in range(0,4):
                if n == m: action_dist[n,m] = 0.
                elif n + m == 1: action_dist[n,m] = 2.
                elif n + m == 5: action_dist[n,m] = 2.
                else: action_dist[n,m] = 0.5

        cost_mat = np.zeros((100,100))
        for i in range(0,25):
            for j in range(0,25):
                base_i, base_j = i*4, j*4
                state_dist = np.linalg.norm(np.array([i//5, i%5]) - np.array([j//5, j%5]))
                cost_mat[base_i:base_i+4,base_j:base_j+4] = state_dist + action_dist
        avg_cost = 1.0
        for i in range(0,2000):
            penalty = init_penalty
            try:
                vis = self.estimate_density_ratio(policy_params)
                avg_cost = np.sum(vis * self.data_dist * self.empirical_cost)
            except: pass
            if avg_cost < (cost_lim - 5e-4):
                break

            for j in range(100):
                optimizer1.zero_grad()
                first_expectation = -(lam * cost_lim) - nu - penalty * torch.sum(torch.multiply(torch.FloatTensor(self.data_dist), torch.multiply(torch.abs(torch.FloatTensor(ref_dist)), f_fn)))
                softmax_policy = torch.nn.functional.softmax(torch.FloatTensor(policy_params), dim=1)
                static_init_bellman = torch.multiply(softmax_policy, value_fn) @ torch.ones((self.num_actions,1))
                bellman_residual = torch.stack([(torch.FloatTensor(self.empirical_transition[s,:,:]) @ static_init_bellman).T for s in range(0, self.num_states)]).squeeze() 
                bellman_residual += (2. * torch.FloatTensor(self.empirical_reward)) - (lam * torch.FloatTensor(self.empirical_cost))
                bellman_residual -= value_fn
                bellman_residual -= nu
                bellman_residual -= penalty * f_fn
                CT = 20.0 * ((2.0 * np.max(self.empirical_cost)) + (penalty * np.max(zeta_fn)))
                objective = -1.0 * (first_expectation - torch.exp(torch.logsumexp(bellman_residual - 1.0, dim=[0,1])))
                objective += torch.nn.functional.relu(torch.abs(torch.stack([f_fn.flatten()]*f_fn.numel()) - f_fn.flatten()[:,None]) - torch.FloatTensor(cost_mat)).sum()
                objective += torch.nn.functional.relu(torch.abs(value_fn) - CT).sum()

                objective.backward()
                optimizer1.step()
            v = value_fn.detach().numpy()
            lv = lam.detach().numpy()
            nv = nu.detach().numpy()


            Q_c = torch.FloatTensor(v)
            lam_value = torch.FloatTensor(np.array([lv]))
            nu_value = torch.FloatTensor(np.array([nv]))
            optimizer.zero_grad()
            softmax_policy = torch.nn.functional.softmax(pi, dim=1)
            init_bellman = torch.mul(softmax_policy, Q_c) @ torch.ones((self.num_actions,1), dtype=torch.float32)
            bellman_residual = (torch.stack([(torch.FloatTensor(self.empirical_transition[s,:,:]) @ init_bellman).T for s in range(0, self.num_states)], dim=1)).squeeze() - Q_c
            bellman_residual += (torch.FloatTensor(self.empirical_reward)) - (lam_value * torch.FloatTensor(self.empirical_cost))
            args = (bellman_residual - nu_value)
            reference_dist = torch.mul(torch.FloatTensor(self.data_dist), torch.abs(torch.FloatTensor(ref_dist)))
            expectation = -1.0 * torch.sum(torch.mul(reference_dist, args))
            policy_obj = (expectation - (lam_value * cost_lim) - nu_value)
            policy_obj.backward()
            optimizer.step()
            policy_params = pi.detach().numpy()
        return policy_params, zeta_fn

    def cost_projection_wass(self, policy_params, ref_dist, init_penalty, cost_lim):
        policy = policy_params
        zeta_fn = np.zeros((self.num_states,self.num_actions))
        lam_value = np.array([0.0])
        nu_value = np.array([0.0])

        v = np.zeros((self.num_states,self.num_actions))
        lv = 0.0
        nv = 0.00
        f_norm = 0.0
        lr = 0.1
        pi = torch.tensor(policy.astype(np.float32), requires_grad=True)
        optimizer = torch.optim.Adam([pi], lr=lr)#, momentum=0.9)

        action_dist = np.zeros((4,4))
        for n in range(0,4):
            for m in range(0,4):
                if n == m: action_dist[n,m] = 0.
                elif n + m == 1: action_dist[n,m] = 2.
                elif n + m == 5: action_dist[n,m] = 2.
                else: action_dist[n,m] = 0.5

        cost_mat = np.zeros((100,100))
        for i in range(0,25):
            for j in range(0,25):
                base_i, base_j = i*4, j*4
                state_dist = np.linalg.norm(np.array([i//5, i%5]) - np.array([j//5, j%5]))
                cost_mat[base_i:base_i+4,base_j:base_j+4] = state_dist + action_dist

        for i in range(0,2000):
            penalty = init_penalty
            ratio = self.estimate_density_ratio(policy_params)
            vis = ratio * self.data_dist
            avg_cost = (vis * self.empirical_cost).sum()
            if avg_cost < (cost_lim):
                break

            value_fn = cp.Variable(policy_params.shape)
            lam = cp.Variable(nonneg=True)
            nu = cp.Variable()
            first_expectation = -(lam * cost_lim) - nu
            softmax_policy = special.softmax(policy_params, axis=1)
            static_init_bellman = cp.multiply(softmax_policy, value_fn) @ np.ones((self.num_actions,1))
            bellman_residual = cp.vstack([(self.empirical_transition[s,:,:] @ static_init_bellman).T for s in range(0, self.num_states)])  - (lam * self.empirical_cost)
            bellman_residual -= value_fn
            f = (bellman_residual - nu) / penalty
            objective = first_expectation - penalty * cp.sum(cp.multiply(self.data_dist, cp.multiply(cp.abs(ref_dist), f)))

            # need to enforce lipschitz constraint
            CT = 50. * ((2.0 * cp.max(self.empirical_cost)) + (penalty * cp.max(zeta_fn)))
            q_lp = cp.Problem(cp.Maximize(objective),
                              [value_fn <= CT,
                              value_fn >= -CT,
                              cp.abs(cp.vstack([f.flatten()]*f.size) - f.flatten()[:,None]) <= cost_mat])
            q_lp.solve()
            try: q_lp.solve()
            except: pass
            if value_fn.value is not None:
                v = value_fn.value
                lv = lam.value
                nv = nu.value
                f_norm = np.linalg.norm((bellman_residual.value - nu.value) / penalty - f.value)

            Q_c = torch.FloatTensor(v)
            lam_value = torch.FloatTensor(np.array([lv]))
            nu_value = torch.FloatTensor(np.array([nv]))
            optimizer.zero_grad()
            softmax_policy = torch.nn.functional.softmax(pi, dim=1)
            init_bellman = torch.mul(softmax_policy, Q_c) @ torch.ones((self.num_actions,1), dtype=torch.float32)
            bellman_residual = (torch.stack([(torch.FloatTensor(self.empirical_transition[s,:,:]) @ init_bellman).T for s in range(0, self.num_states)], dim=1)).squeeze() - Q_c
            bellman_residual -= (lam_value * torch.FloatTensor(self.empirical_cost))
            args = (bellman_residual - nu_value)
            reference_dist = torch.mul(torch.FloatTensor(self.data_dist), torch.abs(torch.FloatTensor(ref_dist)))
            expectation = -1.0 * torch.sum(torch.mul(reference_dist, args))
            policy_obj = (expectation - (lam_value * cost_lim) - nu_value)
            policy_obj.backward()
            optimizer.step()
            policy_params = pi.detach().numpy()
        return policy_params, zeta_fn

    def estimate_density_ratio(self, policy_params, ref_dist=None):
        if ref_dist is None:
            ref_dist = np.ones_like(policy_params)
        zeta_fn = np.ones((self.num_states,self.num_actions))
        value_fn = np.zeros((self.num_states,self.num_actions))
        lam_value = np.array([0.0])

        value_fn = cp.Variable((self.num_states, self.num_actions))
        lam = cp.Variable()
        softmax_policy = special.softmax(policy_params, axis=1)
        first_expectation = -lam
        static_init_bellman = cp.multiply(softmax_policy, value_fn) @ np.ones((self.num_actions,1))
        bellman_residual = cp.vstack([(self.empirical_transition[s,:,:] @ static_init_bellman).T for s in range(0, self.num_states)])
        bellman_residual -= value_fn
        objective = first_expectation + cp.sum(cp.multiply(self.data_dist, cp.multiply(ref_dist, 0.5 * cp.power(lam + bellman_residual, 2.0))))

        CT = 50. * (2.0 * cp.max(self.empirical_cost))
        q_lp = cp.Problem(cp.Minimize(objective),
                            [value_fn <= CT,
                            value_fn >= -CT])
        q_lp.solve()

        zeta = cp.Variable((self.num_states, self.num_actions))
        lam_value = lam.value
        value_fn = value_fn.value
        static_init_bellman = cp.multiply(softmax_policy, value_fn) @ np.ones((self.num_actions,1))
        bellman_residual = cp.vstack([(self.empirical_transition[s,:,:] @ static_init_bellman).T for s in range(0, self.num_states)]) - value_fn
        z_objective = -lam_value + cp.sum(cp.multiply(self.data_dist, cp.multiply(ref_dist, cp.multiply(zeta, lam_value + bellman_residual)))) - cp.sum(cp.multiply(self.data_dist, cp.multiply(ref_dist, 0.5 * cp.power(zeta, 2.0))))
        d_lp = cp.Problem(cp.Maximize(z_objective))
        d_lp.solve()
        return zeta.value

def runEnv(env, policy, num_episodes):
    offline_data = []
    dice_data = []
    expected_return = np.zeros(num_episodes)
    expected_cost = np.zeros(num_episodes)
    per_step_reward = np.zeros(num_episodes)
    per_step_cost = np.zeros(num_episodes)
    for i in range(num_episodes):
        trajectory = []
        dice_traj = []
        env.reset()
        done =  False
        obs = env.state
        discounted_return = 0.0
        total_cost = 0.0
        step = 0
        while not done:
            action = policy.samplAction(obs)
            next_obs, reward, cost, done = env.step(action)
            trajectory.append(Transition(obs,action,reward,cost,next_obs,done))
            dice_traj.append((obs, action, cost, next_obs))
            obs = next_obs
            discounted_return += ((env.gamma ** step) * reward)
            total_cost += ((env.gamma ** step) * cost)
            step += 1
            if step >= 250: break
        offline_data.append(trajectory)
        dice_traj += [(24,policy.samplAction(obs),0,True) for _ in range(50-step)]
        dice_data.append(dice_traj)
        expected_return[i] = discounted_return
        expected_cost[i] = total_cost
        per_step_reward[i] = discounted_return / np.sum(env.gamma ** np.arange(0,step,1))
        per_step_cost[i] = total_cost / np.sum(env.gamma ** np.arange(0,step,1))
    return offline_data, expected_return, expected_cost, dice_data, per_step_reward, per_step_cost

def runner(seed, idx_beta, env):
    np.random.seed(seed)
    torch.manual_seed(seed)

    idx = idx_beta[0]
    beta = idx_beta[1]

    policy_params = np.random.randn(env.num_states,env.num_actions)
    policy = TabularSoftmax(env.num_states,env.num_actions)
    policy.parameters = np.ones((env.num_states,env.num_actions))
    
    data, _, _, dice_data, _, _ = runEnv(env, policy, 1000)
    opt = TabularCOPO((env.num_states,env.num_actions), data, dice_data, env.gamma, 0.1, beta, lambda x: 0.5 * x**2, steps=1)

    policy_params, value_params, density_ratio, value = opt.run_policy_opt()
    policy.parameters = policy_params


    new_dist = density_ratio * opt.data_dist
    new_dist = np.abs(new_dist) / np.abs(new_dist).sum()
    opt.data_dist = new_dist
    
    directly_estimated_density = opt.estimate_density_ratio(policy_params)
    dist_penalty = 50.0
    proj_policy_params, proj_density_ratio = opt.cost_projection_wass_entr_torch(policy_params, directly_estimated_density, dist_penalty, beta)
    policy.parameters = proj_policy_params
    
    data, e_ret, e_cst, p_ret, p_cst = get_batch(env, policy, 10000)
    return [idx, int(beta >= p_cst), p_ret, p_cst]


if __name__ == '__main__':
    np.set_printoptions(precision=6, suppress=True)

    num_trials = 100
    beta = np.arange(0.4,0.9, 0.1)
    num_betas = beta.shape[0]
    betas = [(i, b) for (i,b) in enumerate(beta)]*num_trials
    envs = [WalkAroundGrid() for _ in range(len(betas))]
    
    p = pool.Pool(20)
    arg_list = zip(np.arange(0,len(betas),1), betas, envs)
    success_list = np.array(p.starmap(runner, tqdm.tqdm(arg_list, total=len(betas))))
    p.close()
    p.join()
    np.savetxt("success_list_copo_reward_FHBP.csv", success_list, delimiter=",")

    
