
import numpy as np
from typing import Tuple
from skeleton import Environment


class WalkAroundGrid(Environment):
    def __init__(self, startState=12, shape=(5, 5)):
        self._name = "WalkAroundGrid"
        self._gamma = 1.0
        self._startState = startState

        self._state = self._startState
        self._action = None
        self._reward = 0
        self._cost = 0
        self._isEnd = False

        self._shape = tuple(shape)
        self._size = self._shape[0] * self._shape[1]

        # define stochasticity
        self._prStay = 0.#1
        self._prRotate = 0.#05
        # dicts mapping actions to the appropriate rotations
        self._rotateLeft = {0: 2, 1: 3, 2: 1, 3: 0}
        self._rotateRight = {0: 3, 1: 2, 2: 0, 3: 1}

        self.num_states = 25
        self.num_actions = 4

    @property
    def name(self) -> str:
        return self._name

    @property
    def reward(self) -> float:
        return self._reward

    @property
    def cost(self) -> float:
        return self._cost

    @property
    def action(self) -> int:
        return self._action

    @property
    def isEnd(self) -> bool:
        return self._isEnd

    @property
    def state(self) -> int:
        # return int(np.where(x==1)[0])
        return int(self._state)

    @property
    def gamma(self) -> float:
        return self._gamma

    def nextState(self, state: int, action: int) -> int:
        noise = np.random.uniform()
        if noise < self._prStay:  # do nothing
            return state
        elif noise < (self._prStay + self._prRotate):
            action = self._rotateLeft[action]
        elif noise < (self._prStay + 2 * self._prRotate):
            action = self._rotateRight[action]

        # simulate taking a step in the environment
        nextState = self._state
        if action == 0:  # move up
            nextState = state - self._shape[1]
        elif action == 1:  # move down
            nextState = state + self._shape[1]
        elif action == 2 and (nextState % self._shape[1] != 0):  # move left
            nextState = state - 1
        elif action == 3 and ((nextState + 1) % self._shape[1] != 0):  # move right
            nextState = state + 1

        if nextState >= 0 and nextState < self._size:
            return nextState
        else:
            return state

    def step(self, action: int) -> Tuple[int, float, bool]:

        nextState = self.nextState(self._state, action)
        self._reward = self.R(int(self._state), action, nextState)
        self._state = nextState
        # self._isEnd = self._state == self._endState
        self._cost = self.C(int(self._state), action, nextState)

        return self.state, self.reward, self.cost, self.isEnd

    def reset(self) -> None:
        self._state = self._startState
        self._action = None
        self._reward = 0
        self._cost = 0
        self._isEnd = False

    def R(self, state: int, action: int, nextState: int) -> float:
        """
        Reward is determined by where you came from and where you land.
        Landing in the exterior while pushing wrench counter-clockwise gives 1.0
        Landing in the interior while pushing wrench counter-clockwise gives 0.5
        Everything else gives 0
        [ 0,  1,  2,  3,  4]
        [ 5,  6,  7,  8,  9]
        [10, 11, 12, 13, 14]
        [15, 16, 17, 18, 19]
        [20, 21, 22, 23, 24]

        """
        if state == 0 and nextState == 5: return 1.0
        if state == 1 and nextState == 0: return 1.0
        if state == 1 and nextState == 6: return 0.5
        if state == 2 and nextState == 1: return 1.0
        if state == 3 and nextState == 2: return 1.0
        if state == 4 and nextState == 3: return 1.0
        if state == 5 and nextState == 10: return 1.0
        if state == 10 and nextState == 15: return 1.0
        if state == 15 and nextState == 20: return 1.0
        if state == 15 and nextState == 16: return 0.5
        if state == 20 and nextState == 21: return 1.0
        if state == 21 and nextState == 22: return 1.0
        if state == 22 and nextState == 23: return 1.0
        if state == 23 and nextState == 24: return 1.0
        if state == 23 and nextState == 18: return 0.5
        if state == 24 and nextState == 19: return 1.0
        if state == 19 and nextState == 14: return 1.0
        if state == 14 and nextState == 9: return 1.0
        if state == 9 and nextState == 4: return 1.0
        if state == 9 and nextState == 8: return 0.5

        if state == 7 and nextState == 6: return 0.5
        if state == 6 and nextState == 11: return 0.5
        if state == 6 and nextState == 5: return 1.0
        if state == 11 and nextState == 16: return 0.5
        if state == 16 and nextState == 17: return 0.5
        if state == 16 and nextState == 21: return 1.0
        if state == 17 and nextState == 18: return 0.5
        if state == 18 and nextState == 13: return 0.5
        if state == 18 and nextState == 19: return 1.0
        if state == 13 and nextState == 8: return 0.5
        if state == 8 and nextState == 7: return 0.5
        if state == 8 and nextState == 3: return 1.0

        return 0.0

    def C(self, state: int, action: int, nextState: int) -> float:
        # if nextState == 0: return 1.0
        if (nextState % 5 == 0) or ((nextState + 1) % 5 == 0): return 1.0
        # if 5 % nextState == 0: return 1.0

        return 0.0